# Copyright 2014 Rackspace Australia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import collections
import concurrent.futures
import logging
import os
import threading

from zuul.lib.logutil import get_annotated_logger
from zuul import model


class ReadOnlyBranchCacheError(RuntimeError):
    pass


class BaseConnection(object, metaclass=abc.ABCMeta):
    """Base class for connections.

    A connection is a shared object that sources, triggers and reporters can
    use to speak with a remote API without needing to establish a new
    connection each time or without having to authenticate each time.

    Multiple instances of the same connection may exist with different
    credentials, for example, thus allowing for different pipelines to operate
    on different Gerrit installations or post back as a different user etc.

    Connections can implement their own public methods. Required connection
    methods are validated by the {trigger, source, reporter} they are loaded
    into. For example, a trigger will likely require some kind of query method
    while a reporter may need a review method."""

    log = logging.getLogger('zuul.BaseConnection')

    def __init__(self, driver, connection_name, connection_config):
        # connection_name is the name given to this connection in zuul.ini
        # connection_config is a dictionary of config_section from zuul.ini for
        # this connection.
        # __init__ shouldn't make the actual connection in case this connection
        # isn't used in the layout.
        self.driver = driver
        self.connection_name = connection_name
        self.connection_config = connection_config
        self.sched = None

    def logEvent(self, event):
        log = get_annotated_logger(self.log, event.zuul_event_id)
        log.debug('Scheduling event from {connection}: {event}'.format(
            connection=self.connection_name,
            event=event))
        try:
            if self.sched.statsd:
                self.sched.statsd.incr(
                    'zuul.event.{driver}.{event}'.format(
                        driver=self.driver.name, event=event.type))
                self.sched.statsd.incr(
                    'zuul.event.{driver}.{connection}.{event}'.format(
                        driver=self.driver.name,
                        connection=self.connection_name,
                        event=event.type))
        except Exception:
            self.log.exception("Exception reporting event stats")

    def onLoad(self, zk_client, component_registry):
        pass

    def onStop(self):
        pass

    def registerScheduler(self, sched) -> None:
        self.sched = sched

    def cleanupCache(self):
        """Clean up the connection cache.

        This allows a connection to perform periodic cleanup actions of
        the cache, e.g. garbage collection.
        """
        pass

    def maintainCache(self, relevant, max_age):
        """Remove stale changes from the cache.

        This lets the user supply a list of change cache keys that are
        still in use.  Anything in our cache that isn't in the supplied
        list and is older than the given max. age (in seconds) should
        be safe to remove from the cache."""
        pass

    def getWebController(self, zuul_web):
        """Return a cherrypy web controller to register with zuul-web.

        :param zuul.web.ZuulWeb zuul_web:
            Zuul Web instance.
        :returns: A `zuul.web.handler.BaseWebController` instance.
        """
        return None

    def getEventQueue(self):
        """Return the event queue for this connection.

        :returns: A `zuul.zk.event_queues.ConnectionEventQueue` instance
                  or None.
        """
        return None

    def validateWebConfig(self, config, connections):
        """Validate web config.

        If there is a fatal error, the method should raise an exception.

        :param config:
           The parsed config object.
        :param zuul.lib.connections.ConnectionRegistry connections:
           Registry of all configured connections.
        """
        return False

    def toDict(self):
        """Return public information about the connection
        """
        return {
            "name": self.connection_name,
            "driver": self.driver.name,
        }


class ZKBranchCacheMixin:
    # Expected to be defined by the connection and to be an instance
    # of BranchCache
    _branch_cache = None
    read_only = False

    @abc.abstractmethod
    def isBranchProtected(self, project_name, branch_name,
                          zuul_event_id):
        """Return if the branch is protected or None if the branch is unknown.

        :param str project_name:
            The name of the project.
        :param str branch_name:
            The name of the branch.
        :param zuul_event_id:
            Unique id associated to the handled event.
        """
        pass

    @abc.abstractmethod
    def _fetchProjectBranches(self, project, required_flags):
        """Perform a remote query to determine the project's branches.

        Connection subclasses should implement this method.

        :param model.Project project:
            The project.
        :param set(BranchFlag) required_flags:
            Which flags need to be valid in the result set.

        :returns: A list of BranchInfo objects
        """

    @abc.abstractmethod
    def _getProjectBranchesRequiredFlags(self, exclude_unprotected,
                                         exclude_locked):
        """Calculate the set of required branch flags needed to filter the
        branches as specified.

        Connection subclasses should implement this method.

        :param bool exclude_unprotected:
            Whether to exclude unprotected branches
        :param bool exclude_locked:
            Whether to exclude locked branches

        :returns: A set of BranchFlag objects
        """

    @abc.abstractmethod
    def _filterProjectBranches(self, branch_infos, exclude_unprotected,
                               exclude_locked):
        """Filter branches according to the specified criteria

        Connection subclasses should implement this method.

        :param list(BranchInfo) branch_infos:
            A list of BranchInfo objects previously supplied by
            _fetchProjectBranches
        :param bool exclude_unprotected:
            Whether to exclude unprotected branches
        :param bool exclude_locked:
            Whether to exclude locked branches

        :returns: A list of BranchInfo objects
        """

    def _fetchProjectMergeModes(self, project):
        """Perform a remote query to determine the project's supported merge
           modes.

        Connection subclasses should implement this method if they are
        able to determine which merge modes apply for a project.  The
        default implemantion returns that all merge modes are valid.

        :param model.Project project:
            The project.

        :returns: A list of merge modes as model IDs.

        """
        return model.ALL_MERGE_MODES

    def _fetchProjectDefaultBranch(self, project):
        """Perform a remote query to determine the project's default branch.

        Connection subclasses should implement this method if they are
        able to determine the upstream default branch for a project.  The
        default implemantion returns 'master' for now and will likely change
        to return something else if and when the git default changes.

        :param model.Project project:
            The project.

        :returns: The name of the default branch.

        """
        return 'master'

    def clearConnectionCacheOnBranchEvent(self, event):
        """Update event and clear connection cache if needed.

        This checks whether the event created or deleted a branch so
        that Zuul may know to perform a reconfiguration on the
        project. Drivers must call this method when a branch event is
        received.

        :param event:
            The event, inherit from `zuul.model.TriggerEvent` class.
        """
        if event.oldrev == '0' * 40:
            event.branch_created = True
        elif event.newrev == '0' * 40:
            event.branch_deleted = True
        else:
            event.branch_updated = True

        project = self.source.getProject(event.project_name)
        if event.branch:
            if event.branch_deleted:
                # We currently cannot determine if a deleted branch was
                # protected so we need to assume it was. GitHub/GitLab don't
                # allow deletion of protected branches but we don't get a
                # notification about branch protection settings. Thus we don't
                # know if branch protection has been disabled before deletion
                # of the branch.
                # FIXME(tobiash): Find a way to handle that case
                self.updateProjectBranches(project, event)
            elif event.branch_created:
                # In GitHub, a new branch never can be protected
                # because that needs to be configured after it has
                # been created.  Other drivers could optimize this,
                # but for the moment, implement the lowest common
                # denominator and clear the cache so that we query.
                self.updateProjectBranches(project, event)
            event.branch_cache_ltime = self._branch_cache.ltime
        return event

    def updateProjectBranches(self, project, zuul_event_id=None):
        """Update the branch cache for the project.

        :param zuul.model.Project project:
            The project for which the branches are returned.
        :param zuul_event_id:
            Unique id associated to the handled event.
        """
        log = get_annotated_logger(self.log, event=zuul_event_id)

        # Figure out which queries we have a cache for
        required_flags = self._branch_cache.getProjectCompletedFlags(
            project.name)
        if required_flags:
            # Update them if we have them
            valid_flags, branch_infos = self._fetchProjectBranches(
                project, required_flags)
        else:
            valid_flags = None
            branch_infos = None
        merge_modes = self._fetchProjectMergeModes(project)
        default_branch = self._fetchProjectDefaultBranch(project)
        self._branch_cache.setAllProjectData(
            project.name, valid_flags, branch_infos,
            merge_modes, default_branch)
        log.info("Updated branches for %s", project.name)

    def getProjectBranches(self, project, tenant, min_ltime=-1):
        """Get the branch names for the given project.

        :param zuul.model.Project project:
            The project for which the branches are returned.
        :param zuul.model.Tenant tenant:
            The related tenant.
        :param int min_ltime:
            The minimum ltime to determine if we need to refresh the cache.

        :returns: The list of branch names.
        """
        exclude_unprotected = tenant.getExcludeUnprotectedBranches(project)
        exclude_locked = tenant.getExcludeLockedBranches(project)
        branches = None

        required_flags = self._getProjectBranchesRequiredFlags(
            exclude_unprotected, exclude_locked)
        if self._branch_cache:
            try:
                branches = self._branch_cache.getProjectBranches(
                    project.name, required_flags, min_ltime)
                if branches is not None:
                    branches = [b.name for b in self._filterProjectBranches(
                        branches, exclude_unprotected, exclude_locked)]
            except LookupError:
                if self.read_only:
                    # A scheduler hasn't attempted to fetch them yet
                    raise ReadOnlyBranchCacheError(
                        "Will not fetch project branches as read-only is set")
                else:
                    branches = None

        if branches is not None:
            return sorted(branches)
        elif self.read_only:
            # A scheduler has previously attempted a fetch, but got
            # the None due to an error; we can't retry since we're
            # read-only.
            raise RuntimeError(
                "Will not fetch project branches as read-only is set")

        # Above we calculated a set of flags needed to answer the
        # query.  If the fetch below fails, we will mark that set of
        # flags as failed in the ProjectInfo structure.  However, if
        # the fetch below succeeds, it can supply its own set of valid
        # flags that we will record as successful.  This lets the
        # driver indicate that the returned results include more data
        # than strictly necessary (ie, protected+locked and not just
        # protected).
        try:
            valid_flags, branch_infos = self._fetchProjectBranches(
                project, required_flags)
        except Exception:
            # We weren't able to get the branches.  We need to tell
            # future schedulers to try again but tell zuul-web that we
            # tried and failed.  Set the branches to None to indicate
            # that we have performed a fetch and retrieved no data.  Any
            # time we encounter None in the cache, we will try again.
            if self._branch_cache:
                self._branch_cache.setProjectBranches(
                    project.name, required_flags, None)
            raise
        self.log.info("Got branches for %s" % project.name)

        if self._branch_cache:
            self._branch_cache.setProjectBranches(
                project.name, valid_flags, branch_infos)

        branches = [b.name for b in self._filterProjectBranches(
                    branch_infos, exclude_unprotected, exclude_locked)]
        return sorted(branches)

    def getProjectMergeModes(self, project, tenant, min_ltime=-1):
        """Get the merge modes for the given project.

        :param zuul.model.Project project:
            The project for which the merge modes are returned.
        :param zuul.model.Tenant tenant:
            The related tenant.
        :param int min_ltime:
            The minimum ltime to determine if we need to refresh the cache.

        :returns: The list of merge modes by model id.
        """
        merge_modes = None

        if self._branch_cache:
            try:
                merge_modes = self._branch_cache.getProjectMergeModes(
                    project.name, min_ltime)
            except LookupError:
                if self.read_only:
                    # A scheduler hasn't attempted to fetch them yet
                    raise ReadOnlyBranchCacheError(
                        "Will not fetch merge modes as read-only is set")
                else:
                    merge_modes = None

        if merge_modes is not None:
            return merge_modes
        elif self.read_only:
            # A scheduler has previously attempted a fetch, but got
            # the None due to an error; we can't retry since we're
            # read-only.
            raise RuntimeError(
                "Will not fetch merge_modes as read-only is set")

        # We need to perform a query
        try:
            merge_modes = self._fetchProjectMergeModes(project)
        except Exception:
            # We weren't able to get the merge modes.  We need to tell
            # future schedulers to try again but tell zuul-web that we
            # tried and failed.  Set the merge modes to None to indicate
            # that we have performed a fetch and retrieved no data.  Any
            # time we encounter None in the cache, we will try again.
            if self._branch_cache:
                self._branch_cache.setProjectMergeModes(
                    project.name, None)
            raise
        self.log.info("Got merge modes for %s" % project.name)

        if self._branch_cache:
            self._branch_cache.setProjectMergeModes(
                project.name, merge_modes)

        return merge_modes

    def getProjectDefaultBranch(self, project, tenant, min_ltime=-1):
        """Get the default branch for the given project.

        :param zuul.model.Project project:
            The project for which the default branch is returned.
        :param zuul.model.Tenant tenant:
            The related tenant.
        :param int min_ltime:
            The minimum ltime to determine if we need to refresh the cache.

        :returns: The name of the default branch.
        """
        default_branch = None

        if self._branch_cache:
            try:
                default_branch = self._branch_cache.getProjectDefaultBranch(
                    project.name, min_ltime)
            except LookupError:
                if self.read_only:
                    # A scheduler hasn't attempted to fetch it yet
                    raise ReadOnlyBranchCacheError(
                        "Will not fetch default branch as read-only is set")
                else:
                    default_branch = None

        if default_branch is not None:
            return default_branch
        elif self.read_only:
            # A scheduler has previously attempted a fetch, but got
            # the None due to an error; we can't retry since we're
            # read-only.
            raise RuntimeError(
                "Will not fetch default branch as read-only is set")

        # We need to perform a query
        try:
            default_branch = self._fetchProjectDefaultBranch(project)
        except Exception:
            # We weren't able to get the default branch.  We need to tell
            # future schedulers to try again but tell zuul-web that we
            # tried and failed.  Set the default branch to None to indicate
            # that we have performed a fetch and retrieved no data.  Any
            # time we encounter None in the cache, we will try again.
            if self._branch_cache:
                self._branch_cache.setProjectDefaultBranch(
                    project.name, None)
            raise
        self.log.info("Got default branch for %s: %s", project.name,
                      default_branch)

        if self._branch_cache:
            self._branch_cache.setProjectDefaultBranch(
                project.name, default_branch)

        return default_branch

    def checkBranchCache(self, project_name, event, protected=None):
        """Update the cache for a project when a branch event is processed

        This method must be called when a branch event is processed: if the
        event references a branch and the unprotected branches are excluded,
        the branch protection status could have been changed.

        :params str project_name:
            The project name.
        :params event:
            The event, inherit from `zuul.model.TriggerEvent` class.
        :params protected:
            If defined the caller already knows if the branch is protected
            so the query can be skipped.
        """
        if protected is None:
            protected = self.isBranchProtected(project_name, event.branch,
                                               zuul_event_id=event)
        if protected is not None:
            required_flags = self._getProjectBranchesRequiredFlags(
                exclude_unprotected=True, exclude_locked=False)

            branches = self._branch_cache.getProjectBranches(
                project_name, required_flags, default=None)

            if not branches:
                branches = []

            branches = [b.name for b in branches]

            update = False
            if (event.branch in branches) and (not protected):
                update = True
            if (event.branch not in branches) and (protected):
                update = True
            if update:
                self.log.info("Project %s branch %s protected state "
                              "changed to %s",
                              project_name, event.branch, protected)
                self._branch_cache.setProtected(project_name, event.branch,
                                                protected)
                event.branch_cache_ltime = self._branch_cache.ltime

            event.branch_protected = protected
        else:
            # This can happen if the branch was deleted in
            # GitHub/GitLab. In this case we assume that the branch
            # COULD have been protected before. However, if the event
            # says the branch was created we probably never loaded any
            # configuration from it. The cache update is handled by the
            # push event, so we don't touch the cache here again.
            event.branch_protected = not event.branch_created

    def clearBranchCache(self, projects=None):
        """Clear the branch cache

        In case the branch cache gets out of sync with the source,
        this method can be called to clear it and force querying the
        source the next time the cache is used.
        """
        self._branch_cache.clear(projects)


class ZKChangeCacheMixin:
    # Expected to be defined by the connection and to be an instance
    # that implements the AbstractChangeCache API.
    _change_cache = None

    def cleanupCache(self):
        self._change_cache.cleanup()

    def maintainCache(self, relevant, max_age):
        self._change_cache.prune(relevant, max_age)

    def updateChangeAttributes(self, change, **attrs):
        def _update_attrs(c):
            for name, value in attrs.items():
                setattr(c, name, value)
        self._change_cache.updateChangeWithRetry(change.cache_stat.key,
                                                 change, _update_attrs)

    def estimateCacheDataSize(self):
        return self._change_cache.estimateDataSize()


class BaseThreadPoolEventConnector:
    """Move events from the connection into the scheduler"""

    log = logging.getLogger("zuul.BaseThreadPoolEventConnector")

    def __init__(self, connection):
        self.connection = connection
        self.event_queue = connection.event_queue
        self._stopped = False
        self._events_in_progress = set()
        self._dispatcher_wake_event = threading.Event()
        self._event_dispatcher = threading.Thread(
            name=f'{self.connection.connection_name.title()}EventDispatcher',
            target=self.run_event_dispatcher,
            daemon=True)
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=min(32, (os.cpu_count() or 1) * 4))
        self._event_forward_queue = collections.deque()

    def stop(self):
        self._stopped = True
        self._dispatcher_wake_event.set()
        self.event_queue.election.cancel()
        self._event_dispatcher.join()

        self._thread_pool.shutdown()

    def start(self):
        self._event_dispatcher.start()

    def _shouldStop(self):
        return (self._stopped or
                not self.event_queue.election.is_still_valid())

    def _onNewEvent(self):
        self._dispatcher_wake_event.set()
        # Stop the data watch in case the connector was stopped
        return not self._stopped

    def run_event_dispatcher(self):
        # Wait for the scheduler to prime its config so that we have
        # the full tenant list before we start moving events.
        self.connection.sched.primed_event.wait()
        if self._stopped:
            return
        self.event_queue.registerEventWatch(self._onNewEvent)
        # Set the wake event so we get an initial run
        self._dispatcher_wake_event.set()
        while not self._stopped:
            try:
                self.event_queue.election.run(self._dispatchEventsMain)
            except Exception:
                self.log.exception("Exception handling %s %s event:",
                                   self.connection.connection_name,
                                   self.connection.driver_name)
            # In case we caught an exception with events in progress,
            # reset these in case we run the loop again.
            self._events_in_progress = set()
            self._event_forward_queue = collections.deque()

    def _dispatchEventsMain(self):
        while True:
            delay = None
            if self._dispatcher_wake_event.is_set():
                self._dispatcher_wake_event.clear()
                # We can start processing events as long as we're
                # running; if we are stopping, then we need to continue
                # this loop until previously processed events are
                # completed but not start processing any new events.
                if not self._shouldStop():
                    delay = self._dispatchEvents()

            # Now process the futures from this or any previous
            # iterations of the loop.
            if len(self._event_forward_queue):
                self._forwardEvents()

            # If there are no futures, we can sleep until there are
            # new events (or stop altogether); otherwise we need to
            # continue processing futures.
            if not len(self._event_forward_queue):
                if self._shouldStop():
                    return
                # Delay can be None and we will wait for the next event
                self._dispatcher_wake_event.wait(delay)
            else:
                # Sleep a small amount of time to give the futures
                # time to complete.
                self._dispatcher_wake_event.wait(0.1)
            if delay is not None:
                # If we were asked to delay, then we haven't processed
                # all the events yet so go again.
                self._dispatcher_wake_event.set()

    def _dispatchEvents(self):
        # This is the first half of the event dispatcher.  It reads
        # events from the webhook event queue and passes them to a
        # concurrent executor for pre-processing.
        try:
            event_id_offset = max(
                self.event_queue.eventIdFromAckRef(r)
                for r in self._events_in_progress)
        except ValueError:
            event_id_offset = None

        for event in self.event_queue.iter(event_id_offset):
            if self._shouldStop():
                break

            processor = self._getEventProcessor(event)
            future = self._thread_pool.submit(processor)

            # Events are acknowledged in the event forwarder loop after
            # pre-processing. This way we can ensure that no events are
            # lost.
            self._events_in_progress.add(event.ack_ref)
            self._event_forward_queue.append(future)

    @abc.abstractmethod
    def _getEventProcessor(self, event):
        """Returns a function that will be submitted to the TPE."""

    def _forwardEvents(self):
        # This is the second half of the event dispatcher.  It
        # collects pre-processed events from the concurrent executor
        # and forwards them to the scheduler queues.
        while True:
            try:
                if not len(self._event_forward_queue):
                    return
                # Peek at the next event and see if it's done or if we
                # need to wait for the next loop iteration.
                if not self._event_forward_queue[0].done():
                    return
                future = self._event_forward_queue.popleft()
                (events, connection_event, zuul_event_id,
                 event_type) = future.result()
                try:
                    if not events:
                        log = get_annotated_logger(self.log, zuul_event_id)
                        log.debug('Nothing to be forwarded, '
                                  'event type: %s', event_type)
                    for event in events:
                        self.connection.logEvent(event)
                        if isinstance(event, model.DequeueEvent):
                            self.connection.sched.addChangeManagementEvent(
                                event)
                        else:
                            self.connection.sched.addTriggerEvent(
                                self.connection.driver_name, event
                            )
                finally:
                    # Ack event in Zookeeper
                    self.event_queue.ack(connection_event)
                    self._events_in_progress.remove(connection_event.ack_ref)
            except Exception:
                self.log.exception("Exception moving %s %s event:",
                                   self.connection.connection_name,
                                   self.connection.driver_name)



#!/usr/bin/env python

tenant_config = """---
- tenant:
    name: tenant-one
    source:
      gerrit:
        config-projects:
          - common-config
        untrusted-projects:
          - org/project1
          - org/project2
"""

print(tenant_config)



# Copyright 2018 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import voluptuous as v
from zuul.trigger import BaseTrigger
from zuul.driver.pagure.paguremodel import PagureEventFilter
from zuul.driver.util import scalar_or_list, to_list, make_regex, ZUUL_REGEX


class PagureTrigger(BaseTrigger):
    name = 'pagure'
    log = logging.getLogger("zuul.trigger.PagureTrigger")

    def getEventFilters(self, connection_name, trigger_config,
                        parse_context):
        efilters = []
        pcontext = parse_context
        for trigger in to_list(trigger_config):
            with pcontext.confAttr(trigger, 'event') as attr:
                types = [make_regex(x, pcontext)
                         for x in to_list(attr)]
            with pcontext.confAttr(trigger, 'ref') as attr:
                refs = [make_regex(x, pcontext)
                        for x in to_list(attr)]
            with pcontext.confAttr(trigger, 'comment') as attr:
                comments = [make_regex(x, pcontext)
                            for x in to_list(attr)]

            f = PagureEventFilter(
                connection_name=connection_name,
                trigger=self,
                types=types,
                actions=to_list(trigger.get('action')),
                refs=refs,
                comments=comments,
                statuses=to_list(trigger.get('status')),
                tags=to_list(trigger.get('tag')),
                debug=trigger.get('debug'),
            )
            efilters.append(f)

        return efilters

    def onPullRequest(self, payload):
        pass


def getSchema():
    pagure_trigger = {
        v.Required('event'):
            # Cannot use same event type than github as it collapse
            # with Registered github triggers if any. The Event filter
            # does not have the connections info like the Ref filter (require)
            # have. See manager/__init__.py:addChange
            scalar_or_list(v.Any('pg_pull_request',
                                 'pg_pull_request_review',
                                 'pg_push')),
        'action': scalar_or_list(str),
        'ref': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'comment': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'status': scalar_or_list(str),
        'tag': scalar_or_list(str),
        'debug': bool,
    }

    return pagure_trigger



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import socket

from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPConnection


class ZuulHTTPAdapter(HTTPAdapter):
    "A requests HTTPAdapter class which supports TCP keepalives"
    def __init__(self, *args, **kw):
        self.keepalive = int(kw.pop('keepalive', 0))
        super().__init__(*args, **kw)

    def init_poolmanager(self, *args, **kw):
        if self.keepalive:
            ka = self.keepalive
            kw['socket_options'] = HTTPConnection.default_socket_options + [
                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, ka),
                (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, ka),
                (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 2),
            ]
        return super().init_poolmanager(*args, **kw)



# Copyright 2017 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import time
import urllib.parse
import dateutil.parser
import json

from zuul.model import EventFilter, RefFilter
from zuul.model import Change, TriggerEvent, FalseWithReason
from zuul.driver.util import time_to_seconds, to_list, make_regex
from zuul import exceptions
from zuul.zk.change_cache import ChangeKey


EMPTY_GIT_REF = '0' * 40  # git sha of all zeros, used during creates/deletes


class GerritChange(Change):
    def __init__(self, project):
        super(GerritChange, self).__init__(project)
        self.id = None
        self.status = None
        self.wip = None
        self.approvals = []
        self.missing_labels = set()
        self.submit_requirements = []
        self.commit = None
        self.hashtags = []
        self.zuul_query_ltime = None

    def update(self, data, extra, connection):
        if not (self.zuul_query_ltime and
                self.zuul_query_ltime > data.zuul_query_ltime):
            # If our query is older than the existing change data, it
            # means that another scheduler raced us and won.  Skip the
            # update to avoid having is_merged flap.
            self.zuul_query_ltime = data.zuul_query_ltime
            if data.format == data.SSH:
                self.updateFromSSH(data.data, connection)
            else:
                self.updateFromHTTP(data.data, data.files,
                                    data.commentable_files,
                                    connection)
            for k, v in extra.items():
                setattr(self, k, v)
        key = ChangeKey(connection.connection_name, None,
                        'GerritChange', str(self.number), str(self.patchset))
        return key

    def serialize(self):
        d = super().serialize()
        d.update({
            "id": self.id,
            "status": self.status,
            "wip": self.wip,
            "approvals": self.approvals,
            "missing_labels": list(self.missing_labels),
            "submit_requirements": self.submit_requirements,
            "commit": self.commit,
            "hashtags": self.hashtags,
            "zuul_query_ltime": self.zuul_query_ltime,
        })
        return d

    def deserialize(self, data):
        super().deserialize(data)
        self.id = data.get("id")
        self.status = data["status"]
        self.wip = data["wip"]
        self.approvals = data["approvals"]
        self.missing_labels = set(data["missing_labels"])
        self.submit_requirements = data.get("submit_requirements", [])
        self.commit = data.get("commit")
        self.hashtags = data.get("hashtags", [])
        self.zuul_query_ltime = data.get("zuul_query_ltime")

    def updateFromSSH(self, data, connection):
        if self.patchset is None:
            self.patchset = str(data['currentPatchSet']['number'])
        if 'project' not in data:
            raise exceptions.ChangeNotFound(self.number, self.patchset)
        self.project = connection.source.getProject(data['project'])
        self.commit_id = str(data['currentPatchSet']['revision'])
        self.branch = data['branch']
        self.url = data['url']
        urlparse = urllib.parse.urlparse(connection.baseurl)
        baseurl = "%s://%s%s" % (urlparse.scheme, urlparse.netloc,
                                 urlparse.path)
        baseurl = baseurl.rstrip('/')
        self.uris = [
            '%s/%s' % (baseurl, self.number),
            '%s/#/c/%s' % (baseurl, self.number),
            '%s/c/%s/+/%s' % (baseurl, self.project.name, self.number),
        ]

        max_ps = 0
        files = []
        for ps in data['patchSets']:
            if str(ps['number']) == self.patchset:
                self.ref = ps['ref']
                self.commit = ps['revision']
                # SSH queries gives us a list of dicts for the files. We
                # convert that to a list of filenames.
                for f in ps.get('files', []):
                    files.append(f['file'])
            if int(ps['number']) > int(max_ps):
                max_ps = str(ps['number'])
        if max_ps == self.patchset:
            self.is_current_patchset = True
        else:
            self.is_current_patchset = False
        if len(data.get('parents', [])) > 1:
            # This is a merge commit, and the SSH query only reports
            # files in this commit's content (not files changed by the
            # underlying merged changes).  Set files to None to
            # instruct Zuul to ask the mergers to get the full file
            # list.
            self.files = None
        self.files = files
        self.id = data['id']
        self.is_merged = data.get('status', '') == 'MERGED'
        self.approvals = data['currentPatchSet'].get('approvals', [])
        self.open = data['open']
        self.status = data['status']
        self.wip = data.get('wip', False)
        self.owner = data['owner'].get('username')
        self.message = data['commitMessage']
        self.topic = data.get('topic')
        self.hashtags = data.get('hashtags', [])

        self.missing_labels = set()
        for sr in data.get('submitRecords', []):
            if sr['status'] == 'NOT_READY':
                for label in sr['labels']:
                    if label['status'] in ['OK', 'MAY']:
                        continue
                    elif label['status'] in ['NEED', 'REJECT']:
                        self.missing_labels.add(label['label'])

    def updateFromHTTP(self, data, files, commentable_files, connection):
        urlparse = urllib.parse.urlparse(connection.baseurl)
        baseurl = "%s://%s%s" % (urlparse.scheme, urlparse.netloc,
                                 urlparse.path)
        baseurl = baseurl.rstrip('/')
        current_revision = data['revisions'][data['current_revision']]
        if self.patchset is None:
            self.patchset = str(current_revision['_number'])
        self.project = connection.source.getProject(data['project'])
        self.commit_id = str(data['current_revision'])
        self.id = data['change_id']
        self.branch = data['branch']
        url1 = '%s/%s' % (baseurl, self.number)
        url2 = '%s/#/c/%s' % (baseurl, self.number)
        url3 = '%s/c/%s/+/%s' % (baseurl, self.project.name, self.number)
        if connection.version >= (2, 15, 0):
            # This is what modern Gerrit emits, and if a user
            # navigates to it without access, it provides the best
            # error messages and UX for logging in.
            self.url = url3
        else:
            self.url = url1
        self.uris = [url1, url2, url3]

        for rev_commit, revision in data['revisions'].items():
            if str(revision['_number']) == self.patchset:
                self.ref = revision['ref']
                self.commit = rev_commit

        if str(current_revision['_number']) == self.patchset:
            self.is_current_patchset = True
        else:
            self.is_current_patchset = False

        # HTTP queries give us a dict of files in the form of
        # {filename: { attrs }}. We only want a list of filenames here.
        if files:
            self.files = list(files.keys())
        else:
            self.files = []

        if commentable_files is not None:
            self.commentable_files = list(commentable_files.keys())
        else:
            self.commentable_files = None

        self.is_merged = data['status'] == 'MERGED'
        self.approvals = []
        self.missing_labels = set()
        for label_name, label_data in data.get('labels', {}).items():
            for app in label_data.get('all', []):
                if app.get('value', 0) == 0:
                    continue
                by = {}
                for k in ('name', 'username', 'email'):
                    if k in app:
                        by[k] = app[k]
                self.approvals.append({
                    "type": label_name,
                    "description": label_name,
                    "value": app['value'],
                    "grantedOn":
                    dateutil.parser.parse(app['date']).timestamp(),
                    "by": by,
                })
            if label_data.get('optional', False):
                continue
            if label_data.get('blocking', False):
                self.missing_labels.add(label_name)
                continue
            if 'approved' in label_data:
                continue
            self.missing_labels.add(label_name)
        self.submit_requirements = data.get('submit_requirements', [])
        self.open = data['status'] == 'NEW'
        self.status = data['status']
        self.wip = data.get('work_in_progress', False)
        self.owner = data['owner'].get('username')
        self.message = current_revision['commit']['message']
        self.topic = data.get('topic')
        self.hashtags = data.get('hashtags', [])


class GerritTriggerEvent(TriggerEvent):
    """Incoming event from an external system."""
    def __init__(self):
        super(GerritTriggerEvent, self).__init__()
        self.approvals = []
        self.uuid = None
        self.scheme = None
        self.patchsetcomments = None
        self.added = None  # Used by hashtags-changed event
        self.removed = None  # Used by hashtags-changed event
        self.default_branch_changed = None

    @classmethod
    def fromGerritEventDict(cls, data, timestamp, connection,
                            zuul_event_id):
        event = cls()
        event.timestamp = timestamp
        event.connection_name = connection.connection_name
        event.zuul_event_id = zuul_event_id

        event.type = data.get('type')
        event.uuid = data.get('uuid')

        # This catches when a change is merged, as it could potentially
        # have merged layout info which will need to be read in.
        # Ideally this would be done with a refupdate event so as to catch
        # directly pushed things as well as full changes being merged.
        # But we do not yet get files changed data for pure refupdate events.
        # TODO(jlk): handle refupdated events instead of just changes
        if event.type == 'change-merged':
            event.branch_updated = True
        event.trigger_name = 'gerrit'
        change = data.get('change')
        event.project_hostname = connection.canonical_hostname
        if change:
            event.project_name = change.get('project')
            event.branch = change.get('branch')
            event.change_number = str(change.get('number'))
            event.change_url = change.get('url')
            patchset = data.get('patchSet')
            if patchset:
                event.patch_number = str(patchset.get('number'))
                event.ref = patchset.get('ref')
            event.approvals = data.get('approvals', [])
            event.comment = data.get('comment')
            patchsetcomments = data.get('patchSetComments', {}).get(
                "/PATCHSET_LEVEL")
            if patchsetcomments:
                event.patchsetcomments = []
                for patchsetcomment in patchsetcomments:
                    event.patchsetcomments.append(
                        patchsetcomment.get('message'))
            event.added = data.get('added')
            event.removed = data.get('removed')
        refupdate = data.get('refUpdate')
        if refupdate:
            event.project_name = refupdate.get('project')
            event.ref = refupdate.get('refName')
            event.oldrev = refupdate.get('oldRev')
            event.newrev = refupdate.get('newRev')
        if event.project_name is None:
            # ref-replica* events
            event.project_name = data.get('project')
        if event.type == 'project-created':
            event.project_name = data.get('projectName')
        if event.type == 'project-head-updated':
            event.project_name = data.get('projectName')
            event.ref = data.get('newHead')
            event.branch = event.ref[len('refs/heads/'):]
            event.default_branch_changed = True
        # Map the event types to a field name holding a Gerrit
        # account attribute. See Gerrit stream-event documentation
        # in cmd-stream-events.html
        accountfield_from_type = {
            'patchset-created': 'uploader',
            'draft-published': 'uploader',  # Gerrit 2.5/2.6
            'change-abandoned': 'abandoner',
            'change-restored': 'restorer',
            'change-merged': 'submitter',
            'merge-failed': 'submitter',  # Gerrit 2.5/2.6
            'comment-added': 'author',
            'ref-updated': 'submitter',
            'reviewer-added': 'reviewer',  # Gerrit 2.5/2.6
            'topic-changed': 'changer',
            'hashtags-changed': 'editor',
            'vote-deleted': 'deleter',
            'project-created': None,  # Gerrit 2.14
            'pending-check': None,  # Gerrit 3.0+
            'project-head-updated': None,
        }
        event.account = None
        event._accountfield_unknown = False
        if event.type in accountfield_from_type:
            field = accountfield_from_type[event.type]
            if field:
                event.account = data.get(accountfield_from_type[event.type])
        else:
            event._accountfield_unknown = True

        # This checks whether the event created or deleted a branch so
        # that Zuul may know to perform a reconfiguration on the
        # project.
        branch_refs = 'refs/heads/'
        event._branch_ref_update = False
        if (event.type == 'ref-updated' and
            ((not event.ref.startswith('refs/')) or
             event.ref.startswith(branch_refs))):

            if event.ref.startswith(branch_refs):
                event.branch = event.ref[len(branch_refs):]
            else:
                event.branch = event.ref

            event._branch_ref_update = True
        return event

    def toDict(self):
        d = super().toDict()
        d["approvals"] = self.approvals
        d["uuid"] = self.uuid
        d["scheme"] = self.scheme
        d["patchsetcomments"] = self.patchsetcomments
        d["added"] = self.added
        d["removed"] = self.removed
        d["default_branch_changed"] = self.default_branch_changed
        return d

    def updateFromDict(self, d):
        super().updateFromDict(d)
        self.approvals = d["approvals"]
        self.uuid = d["uuid"]
        self.scheme = d["scheme"]
        self.patchsetcomments = d["patchsetcomments"]
        self.added = d.get("added")
        self.removed = d.get("removed")
        self.default_branch_changed = d.get("default_branch_changed")

    def __repr__(self):
        ret = '<GerritTriggerEvent %s %s' % (self.type,
                                             self.canonical_project_name)

        if self.branch:
            ret += " %s" % self.branch
        if self.change_number:
            ret += " %s,%s" % (self.change_number, self.patch_number)
        if self.approvals:
            ret += ' ' + ', '.join(
                ['%s:%s' % (a['type'], a['value']) for a in self.approvals])
        if self.added:
            ret += f" added {self.added}"
        if self.removed:
            ret += f" removed {self.removed}"
        ret += '>'

        return ret

    def isPatchsetCreated(self):
        return self.type in ('patchset-created', 'pending-check')

    def isChangeAbandoned(self):
        return 'change-abandoned' == self.type

    def isDefaultBranchChanged(self):
        return bool(self.default_branch_changed)


class GerritEventFilter(EventFilter):
    def __init__(self, connection_name, trigger, types=[], branches=[],
                 refs=[], event_approvals={}, event_approval_changes={},
                 comments=[], emails=[], usernames=[], required_approvals=[],
                 reject_approvals=[], added=[], removed=[], uuid=None,
                 scheme=None, ignore_deletes=True, require=None, reject=None,
                 debug=None, parse_context=None):

        EventFilter.__init__(self, connection_name, trigger, debug)

        # TODO: Backwards compat, remove after 9.x:
        if required_approvals and require is None:
            require = {'approval': required_approvals}
        if reject_approvals and reject is None:
            reject = {'approval': reject_approvals}

        if require:
            self.require_filter = GerritRefFilter.requiresFromConfig(
                connection_name, require, parse_context)
        else:
            self.require_filter = None

        if reject:
            self.reject_filter = GerritRefFilter.rejectFromConfig(
                connection_name, reject, parse_context)
        else:
            self.reject_filter = None

        self._types = [x.pattern for x in types]
        self._branches = [x.pattern for x in branches]
        self._refs = [x.pattern for x in refs]
        self._comments = [x.pattern for x in comments]
        self._emails = [x.pattern for x in emails]
        self._usernames = [x.pattern for x in usernames]
        self._added = [x.pattern for x in added]
        self._removed = [x.pattern for x in removed]
        self.types = types
        self.branches = branches
        self.refs = refs
        self.comments = comments
        self.emails = emails
        self.usernames = usernames
        self.added = added
        self.removed = removed
        self.event_approvals = event_approvals
        self.event_approval_changes = event_approval_changes
        self.uuid = uuid
        self.scheme = scheme
        self.ignore_deletes = ignore_deletes

        self._hash = hash(json.dumps(self.toDict(), sort_keys=True))

    def __hash__(self):
        return self._hash

    def toDict(self):
        if self.require_filter:
            require_filter = self.require_filter.toDict()
        else:
            require_filter = None
        if self.reject_filter:
            reject_filter = self.reject_filter.toDict()
        else:
            reject_filter = None
        return dict(
            require_filter=require_filter,
            reject_filter=reject_filter,
            types=self._types,
            branches=self._branches,
            refs=self._refs,
            comments=self._comments,
            emails=self._emails,
            usernames=self._usernames,
            added=self._added,
            removed=self._removed,
            event_approvals=self.event_approvals,
            event_approval_changes=self.event_approval_changes,
            uuid=self.uuid,
            scheme=self.scheme,
            ignore_deletes=self.ignore_deletes,
        )

    def __eq__(self, other):
        return (
            isinstance(other, GerritEventFilter) and
            self.require_filter == other.require_filter and
            self.reject_filter == other.reject_filter and
            self.types == other.types and
            self.branches == other.branches and
            self.refs == other.refs and
            self.comments == other.comments and
            self.emails == other.emails and
            self.usernames == other.usernames and
            self.added == other.added and
            self.removed == other.removed and
            self.event_approvals == other.event_approvals and
            self.event_approval_changes == other.event_approval_changes and
            self.uuid == other.uuid and
            self.scheme == other.scheme and
            self.ignore_deletes == other.ignore_deletes
        )

    def __repr__(self):
        ret = '<GerritEventFilter'
        ret += ' connection: %s' % self.connection_name

        if self._types:
            ret += ' types: %s' % ', '.join(self._types)
        if self.uuid:
            ret += ' uuid: %s' % (self.uuid,)
        if self.scheme:
            ret += ' scheme: %s' % (self.scheme,)
        if self._branches:
            ret += ' branches: %s' % ', '.join(self._branches)
        if self._refs:
            ret += ' refs: %s' % ', '.join(self._refs)
        if self.ignore_deletes:
            ret += ' ignore_deletes: %s' % self.ignore_deletes
        if self.event_approvals:
            ret += ' event_approvals: %s' % ', '.join(
                ['%s:%s' % a for a in self.event_approvals.items()])
        if self.event_approval_changes:
            ret += ' event_approval_changes: %s' % ', '.join(
                ['%s:%s' % a for a in self.event_approval_changes.items()])
        if self._comments:
            ret += ' comments: %s' % ', '.join(self._comments)
        if self._emails:
            ret += ' emails: %s' % ', '.join(self._emails)
        if self._usernames:
            ret += ' usernames: %s' % ', '.join(self._usernames)
        if self._added:
            ret += ' added: %s' % ', '.join(self._added)
        if self._removed:
            ret += ' removed: %s' % ', '.join(self._removed)
        if self.require_filter:
            ret += ' require: %s' % repr(self.require_filter)
        if self.reject_filter:
            ret += ' reject: %s' % repr(self.reject_filter)
        ret += '>'

        return ret

    def _checkEventType(self, event_type):
        matches_type = False
        for etype in self.types:
            if etype.match(event_type):
                matches_type = True
        if self.types and not matches_type:
            return FalseWithReason("Types %s do not match %s" % (
                self.types, event_type))

    def _checkEventBranch(self, event_branch):
        # branches are ORed
        matches_branch = False
        for branch in self.branches:
            if branch.match(event_branch):
                matches_branch = True
        if self.branches and not matches_branch:
            return FalseWithReason("Branches %s do not match %s" % (
                self.branches, event_branch))

    def _checkEventRef(self, event_ref):
        # refs are ORed
        matches_ref = False
        if event_ref is not None:
            for ref in self.refs:
                if ref.match(event_ref):
                    matches_ref = True
        if self.refs and not matches_ref:
            return FalseWithReason(
                "Refs %s do not match %s" % (self.refs, event_ref))

    def preFilter(self, event):
        # This is specific to the gerrit driver.
        # Perform some quick checks to see if we can ignore this
        # event.  If we don't know that we can ignore it, it will be
        # enqueued and go through the normal matching process.
        r = self._checkEventType(event.type)
        if r is not None:
            return r

        r = self._checkEventBranch(event.branch)
        if r is not None:
            return r

        r = self._checkEventRef(event.ref)
        if r is not None:
            return r

        return True

    def matches(self, event, change):
        if not super().matches(event, change):
            return False

        r = self._checkEventType(event.type)
        if r is not None:
            return r

        if event.type == 'pending-check':
            if self.uuid and event.uuid != self.uuid:
                return False
            if self.scheme and event.uuid.split(':')[0] != self.scheme:
                return False

        r = self._checkEventBranch(event.branch)
        if r is not None:
            return r

        r = self._checkEventRef(event.ref)
        if r is not None:
            return r

        if self.ignore_deletes and event.newrev == EMPTY_GIT_REF:
            # If the updated ref has an empty git sha (all 0s),
            # then the ref is being deleted
            return FalseWithReason("Ref deletion events are ignored")

        # comments are ORed
        matches_comment_re = False
        for comment_re in self.comments:
            if (event.comment is not None and
                comment_re.search(event.comment)):
                matches_comment_re = True
            if event.patchsetcomments is not None:
                for comment in event.patchsetcomments:
                    if (comment is not None and
                        comment_re.search(comment)):
                        matches_comment_re = True
        if self.comments and not matches_comment_re:
            return FalseWithReason("Comments %s do not match %s" % (
                self.comments, event.patchsetcomments))

        # We better have an account provided by Gerrit to do
        # email filtering.
        if event.account is not None:
            account_email = event.account.get('email')
            # emails are ORed
            matches_email_re = False
            for email_re in self.emails:
                if (account_email is not None and
                        email_re.search(account_email)):
                    matches_email_re = True
            if self.emails and not matches_email_re:
                return FalseWithReason("Username %s does not match %s" % (
                    self.emails, account_email))

            # usernames are ORed
            account_username = event.account.get('username')
            matches_username_re = False
            for username_re in self.usernames:
                if (account_username is not None and
                    username_re.search(account_username)):
                    matches_username_re = True
            if self.usernames and not matches_username_re:
                return FalseWithReason("Username %s does not match %s" % (
                    self.usernames, account_username))

        # approvals are ANDed
        for category, value in self.event_approvals.items():
            matches_approval = False
            for eapp in event.approvals:
                if (eapp['description'] == category and
                        int(eapp['value']) == int(value)):
                    matches_approval = True
            if not matches_approval:
                return FalseWithReason("Approvals %s do not match %s" % (
                    self.event_approvals, event.approvals))

        for category, value in self.event_approval_changes.items():
            matches_approval = False
            for eapp in event.approvals:
                if (eapp['description'] == category and
                        int(eapp['value']) == int(value) and
                        'oldValue' in eapp and
                        int(eapp['value']) != int(eapp['oldValue'])):
                    matches_approval = True
            if not matches_approval:
                return FalseWithReason(
                    "Changed approvals %s do not match %s" % (
                        self.event_approval_changes, event.approvals))

        # hashtags are ORed
        if self.added:
            matches_token = False
            event_added = event.added or []
            for action_re in self.added:
                if matches_token:
                    break
                for token in event_added:
                    if action_re.search(token):
                        matches_token = True
                        break
            if not matches_token:
                return FalseWithReason("Added %s does not match %s" % (
                    self.added, event.added))
        if self.removed:
            matches_token = False
            event_removed = event.removed or []
            for action_re in self.removed:
                if matches_token:
                    break
                for token in event_removed:
                    if action_re.search(token):
                        matches_token = True
                        break
            if not matches_token:
                return FalseWithReason("Removed %s does not match %s" % (
                    self.removed, event.removed))

        if self.require_filter:
            require_filter_result = self.require_filter.matches(change)
            if not require_filter_result:
                return require_filter_result

        if self.reject_filter:
            reject_filter_result = self.reject_filter.matches(change)
            if not reject_filter_result:
                return reject_filter_result

        return True


class GerritRefFilter(RefFilter):
    def __init__(self, connection_name,
                 parse_context,
                 open=None, reject_open=None,
                 current_patchset=None, reject_current_patchset=None,
                 wip=None, reject_wip=None,
                 statuses=[], reject_statuses=[],
                 required_approvals=[], reject_approvals=[],
                 required_hashtags=[], reject_hashtags=[]):
        RefFilter.__init__(self, connection_name)

        self._required_approvals = required_approvals
        self.required_approvals = self._tidy_approvals(
            copy.deepcopy(required_approvals), parse_context)
        self._reject_approvals = reject_approvals
        self.reject_approvals = self._tidy_approvals(
            copy.deepcopy(reject_approvals), parse_context)
        self.statuses = statuses
        self.reject_statuses = reject_statuses
        self.required_hashtags = required_hashtags
        self.reject_hashtags = reject_hashtags

        if reject_open is not None:
            self.open = not reject_open
        else:
            self.open = open
        if reject_wip is not None:
            self.wip = not reject_wip
        else:
            self.wip = wip
        if reject_current_patchset is not None:
            self.current_patchset = not reject_current_patchset
        else:
            self.current_patchset = current_patchset

    def toDict(self):
        return dict(
            required_approvals=self._required_approvals,
            reject_approvals=self._reject_approvals,
            statuses=self.statuses,
            reject_statuses=self.reject_statuses,
            required_hashtags=self.required_hashtags,
            reject_hashtags=self.reject_hashtags,
            open=self.open,
            wip=self.wip,
            current_patchset=self.current_patchset,
        )

    def __eq__(self, other):
        return (
            isinstance(other, GerritRefFilter) and
            self.required_approvals == other.required_approvals and
            self.reject_approvals == other.reject_approvals and
            self.statuses == other.statuses and
            self.reject_statuses == other.reject_statuses and
            self.required_hashtags == other.required_hashtags and
            self.reject_hashtags == other.reject_hashtags and
            self.open == other.open and
            self.wip == other.wip and
            self.current_patchset == other.current_patchset
        )

    @classmethod
    def requiresFromConfig(cls, connection_name, config, parse_context):
        with parse_context.confAttr(config, 'hashtags') as attr:
            hashtags = [make_regex(x, parse_context) for x in to_list(attr)]
        return cls(
            connection_name=connection_name,
            parse_context=parse_context,
            open=config.get('open'),
            current_patchset=config.get('current-patchset'),
            wip=config.get('wip'),
            statuses=to_list(config.get('status')),
            required_approvals=to_list(config.get('approval')),
            required_hashtags=hashtags,
        )

    @classmethod
    def rejectFromConfig(cls, connection_name, config, parse_context):
        with parse_context.confAttr(config, 'hashtags') as attr:
            hashtags = [make_regex(x, parse_context) for x in to_list(attr)]
        return cls(
            connection_name=connection_name,
            parse_context=parse_context,
            reject_open=config.get('open'),
            reject_current_patchset=config.get('current-patchset'),
            reject_wip=config.get('wip'),
            reject_statuses=to_list(config.get('status')),
            reject_approvals=to_list(config.get('approval')),
            reject_hashtags=hashtags,
        )

    def __repr__(self):
        ret = '<GerritRefFilter'

        ret += ' connection_name: %s' % self.connection_name
        if self.open is not None:
            ret += ' open: %s' % self.open
        if self.wip is not None:
            ret += ' wip: %s' % self.wip
        if self.current_patchset is not None:
            ret += ' current-patchset: %s' % self.current_patchset
        if self.statuses:
            ret += ' statuses: %s' % ', '.join(self.statuses)
        if self.reject_statuses:
            ret += ' reject-statuses: %s' % ', '.join(self.reject_statuses)
        if self.required_approvals:
            ret += (' required-approvals: %s' %
                    str(self.required_approvals))
        if self.reject_approvals:
            ret += (' reject-approvals: %s' %
                    str(self.reject_approvals))
        if self.required_hashtags:
            ret += (' required-hashtags: %s' %
                    [x.pattern for x in self.required_hashtags])
        if self.reject_hashtags:
            ret += (' reject-hashtags: %s' %
                    [x.pattern for x in self.reject_hashtags])
        ret += '>'

        return ret

    def matches(self, change):
        if self.open is not None:
            # if a "change" has no number, it's not a change, but a push
            # and cannot possibly pass this test.
            if hasattr(change, 'number'):
                if self.open != change.open:
                    return FalseWithReason(
                        "Change does not match open requirement")
            else:
                return FalseWithReason("Ref is not a Change")

        if self.current_patchset is not None:
            # if a "change" has no number, it's not a change, but a push
            # and cannot possibly pass this test.
            if hasattr(change, 'number'):
                if self.current_patchset != change.is_current_patchset:
                    return FalseWithReason(
                        "Change does not match current patchset requirement")
            else:
                return FalseWithReason("Ref is not a Change")

        if self.wip is not None:
            # if a "change" has no number, it's not a change, but a push
            # and cannot possibly pass this test.
            if hasattr(change, 'number'):
                if self.wip != change.wip:
                    return FalseWithReason(
                        "Change does not match WIP requirement")
            else:
                return FalseWithReason("Ref is not a Change")

        if self.statuses:
            if change.status not in self.statuses:
                return FalseWithReason(
                    "Required statuses %s do not match %s" % (
                        self.statuses, change.status))
        if self.reject_statuses:
            if change.status in self.reject_statuses:
                return FalseWithReason(
                    "Reject statuses %s match %s" % (
                        self.reject_statuses, change.status))

        for hashtag_re in self.required_hashtags:
            matches_hashtag = False
            for token in change.hashtags:
                if hashtag_re.search(token):
                    matches_hashtag = True
                    break
            if not matches_hashtag:
                return FalseWithReason(
                    "Required hashtags %s do not match %s" % (
                        [x.pattern for x in self.required_hashtags],
                        change.hashtags))
        for hashtag_re in self.reject_hashtags:
            for token in change.hashtags:
                if hashtag_re.search(token):
                    return FalseWithReason(
                        "Reject hashtags %s match %s" % (
                            [x.pattern for x in self.reject_hashtags],
                            change.hashtags))

        # required approvals are ANDed (reject approvals are ORed)
        matches_approvals_result = self.matchesApprovals(change)
        if not matches_approvals_result:
            return matches_approvals_result

        return True

    def _tidy_approvals(self, approvals, parse_context):
        for a in approvals:
            if 'username' in a:
                with parse_context.confAttr(a, 'username') as v:
                    a['username'] = make_regex(v, parse_context)
            if 'email' in a:
                with parse_context.confAttr(a, 'email') as v:
                    a['email'] = make_regex(v, parse_context)
            if 'newer-than' in a:
                with parse_context.confAttr(a, 'newer-than') as v:
                    a['newer-than'] = time_to_seconds(v)
            if 'older-than' in a:
                with parse_context.confAttr(a, 'older-than') as v:
                    a['older-than'] = time_to_seconds(v)
        return approvals

    def _match_approval_required_approval(self, rapproval, approval):
        # Check if the required approval and approval match
        if 'description' not in approval:
            return False
        now = time.time()
        by = approval.get('by', {})
        for k, v in rapproval.items():
            if k == 'username':
                if (not v.search(by.get('username', ''))):
                    return False
            elif k == 'email':
                if (not v.search(by.get('email', ''))):
                    return False
            elif k == 'newer-than':
                t = now - v
                if (approval['grantedOn'] < t):
                    return False
            elif k == 'older-than':
                t = now - v
                if (approval['grantedOn'] >= t):
                    return False
            else:
                if not isinstance(v, list):
                    v = [v]
                if (approval['description'] != k or
                        int(approval['value']) not in v):
                    return False
        return True

    def matchesApprovals(self, change):
        if self.required_approvals or self.reject_approvals:
            if not hasattr(change, 'number'):
                # Not a change, no reviews
                return FalseWithReason("Ref is not a Change")
        if self.required_approvals and not change.approvals:
            # A change with no approvals can not match
            return FalseWithReason("Approvals %s does not match %s" % (
                self.required_approvals, change.approvals))

        # TODO(jhesketh): If we wanted to optimise this slightly we could
        # analyse both the REQUIRE and REJECT filters by looping over the
        # approvals on the change and keeping track of what we have checked
        # rather than needing to loop on the change approvals twice
        return (self.matchesRequiredApprovals(change) and
                self.matchesNoRejectApprovals(change))

    def matchesRequiredApprovals(self, change):
        # Check if any approvals match the requirements
        for rapproval in self.required_approvals:
            matches_rapproval = False
            for approval in change.approvals:
                if self._match_approval_required_approval(rapproval, approval):
                    # We have a matching approval so this requirement is
                    # fulfilled
                    matches_rapproval = True
                    break
            if not matches_rapproval:
                return FalseWithReason(
                    "Required approvals %s do not match %s" % (
                        self.required_approvals, change.approvals))
        return True

    def matchesNoRejectApprovals(self, change):
        # Check to make sure no approvals match a reject criteria
        for rapproval in self.reject_approvals:
            for approval in change.approvals:
                if self._match_approval_required_approval(rapproval, approval):
                    # A reject approval has been matched, so we reject
                    # immediately
                    return FalseWithReason("Reject approvals %s match %s" % (
                        self.reject_approvals, change.approvals))
        # To get here no rejects can have been matched so we should be good to
        # queue
        return True



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013-2014 OpenStack Foundation
# Copyright 2021-2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import sys
import signal

import zuul.cmd
import zuul.executor.server

from zuul.lib.config import get_default


class Executor(zuul.cmd.ZuulDaemonApp):
    app_name = 'executor'
    app_description = 'A standalone Zuul executor.'

    def createParser(self):
        parser = super(Executor, self).createParser()
        parser.add_argument('--keep-jobdir', dest='keep_jobdir',
                            action='store_true',
                            help='keep local jobdirs after run completes')
        self.addSubCommands(parser, zuul.executor.server.COMMANDS)
        return parser

    def parseArguments(self, args=None):
        super(Executor, self).parseArguments()

    def exit_handler(self, signum, frame):
        if self.config.has_option('executor', 'sigterm_method'):
            graceful = self.config.get('executor', 'sigterm_method')
        else:
            graceful = 'graceful'
        if graceful.lower() == 'graceful':
            self.executor.graceful()
        elif graceful.lower() == 'stop':
            self.executor.stop()
        else:
            self.log.error("Unknown value for executor.sigterm_method:"
                           f"'{graceful}'. Expected 'graceful' or 'stop'")
            self.executor.graceful()

    def start_log_streamer(self):
        pipe_read, pipe_write = os.pipe()
        child_pid = os.fork()
        if child_pid == 0:
            os.close(pipe_write)
            import zuul.lib.log_streamer

            self.log.info("Starting log streamer")
            streamer = zuul.lib.log_streamer.LogStreamer(
                '::', self.finger_port, self.job_dir)

            # Keep running until the parent dies:
            pipe_read = os.fdopen(pipe_read)
            try:
                pipe_read.read()
            except KeyboardInterrupt:
                pass
            self.log.info("Stopping log streamer")
            streamer.stop()
            os._exit(0)
        else:
            os.close(pipe_read)
            self.log_streamer_pipe = pipe_write
            self.log_streamer_pid = child_pid

    def run(self):
        self.handleCommands()

        self.setup_logging('executor', 'log_config')
        self.log = logging.getLogger("zuul.Executor")

        self.configure_connections(sources=True, check_bwrap=True)

        if self.config.has_option('executor', 'job_dir'):
            self.job_dir = os.path.expanduser(
                self.config.get('executor', 'job_dir'))
            if not os.path.isdir(self.job_dir):
                print("Invalid job_dir: {job_dir}".format(
                    job_dir=self.job_dir))
                sys.exit(1)
        else:
            self.job_dir = '/var/lib/zuul/builds'
            if not os.path.exists(self.job_dir):
                os.mkdir(self.job_dir)

        self.finger_port = int(
            get_default(self.config, 'executor', 'finger_port',
                        zuul.executor.server.DEFAULT_FINGER_PORT)
        )

        self.start_log_streamer()

        ExecutorServer = zuul.executor.server.ExecutorServer
        self.executor = ExecutorServer(self.config,
                                       self.connections,
                                       jobdir_root=self.job_dir,
                                       keep_jobdir=self.args.keep_jobdir,
                                       log_streaming_port=self.finger_port)
        self.executor.start()

        if self.args.nodaemon:
            signal.signal(signal.SIGTERM, self.exit_handler)

        while True:
            try:
                self.executor.join()
                break
            except KeyboardInterrupt:
                print("Ctrl + C: asking executor to exit nicely...\n")
                self.exit_handler(signal.SIGINT, None)

        os.close(self.log_streamer_pipe)
        os.waitpid(self.log_streamer_pid, 0)


def main():
    Executor().main()



# Copyright 2022, 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os

from zuul.zk.components import (
    ComponentRegistry,
)
from tests.base import (
    ZuulTestCase,
    gerrit_config,
    simple_layout,
    iterate_timeout,
    model_version,
    FIXTURE_DIR,
)
from zuul import model
from zuul.zk.locks import management_queue_lock


class TestModelUpgrade(ZuulTestCase):
    tenant_config_file = "config/single-tenant/main-model-upgrade.yaml"
    scheduler_count = 1

    def getJobData(self, tenant, pipeline):
        item_path = f'/zuul/tenant/{tenant}/pipeline/{pipeline}/item'
        count = 0
        for item in self.zk_client.client.get_children(item_path):
            bs_path = f'{item_path}/{item}/buildset'
            for buildset in self.zk_client.client.get_children(bs_path):
                data = json.loads(self.getZKObject(
                    f'{bs_path}/{buildset}/job/check-job'))
                count += 1
                yield data
        if not count:
            raise Exception("No job data found")

    @model_version(0)
    @simple_layout('layouts/simple.yaml')
    def test_model_upgrade_0_1(self):
        component_registry = ComponentRegistry(self.zk_client)
        self.assertEqual(component_registry.model_api, 0)

        # Upgrade our component
        self.model_test_component_info.model_api = 1

        for _ in iterate_timeout(30, "model api to update"):
            if component_registry.model_api == 1:
                break

    @model_version(33)
    def test_model_upgrade_33_34(self):

        attrs = model.SystemAttributes.fromDict({
            "use_relative_priority": True,
            "max_hold_expiration": 7200,
            "default_hold_expiration": 3600,
            "default_ansible_version": "X",
            "web_root": "/web/root",
            "websocket_url": "/web/socket",
            "web_status_url": "ignored",
        })

        attr_dict = attrs.toDict()
        self.assertIn("web_status_url", attr_dict)
        self.assertEqual(attr_dict["web_status_url"], "")

        # Upgrade our component
        self.model_test_component_info.model_api = 34

        component_registry = ComponentRegistry(self.zk_client)
        for _ in iterate_timeout(30, "model api to update"):
            if component_registry.model_api == 34:
                break

        attr_dict = attrs.toDict()
        self.assertNotIn("web_status_url", attr_dict)


class TestModelUpgradeGerritCircularDependencies(ZuulTestCase):
    config_file = "zuul-gerrit-github.conf"
    tenant_config_file = "config/circular-dependencies/main.yaml"

    @model_version(31)
    @gerrit_config(submit_whole_topic=True)
    def test_model_31_32(self):
        self.executor_server.hold_jobs_in_build = True

        A = self.fake_gerrit.addFakeChange('org/project1', "master", "A",
                                           topic='test-topic')
        B = self.fake_gerrit.addFakeChange('org/project2', "master", "B",
                                           topic='test-topic')

        A.addApproval("Code-Review", 2)
        B.addApproval("Code-Review", 2)
        B.addApproval("Approved", 1)

        self.fake_gerrit.addEvent(A.addApproval("Approved", 1))
        self.waitUntilSettled()

        first = self.scheds.first
        second = self.createScheduler()
        second.start()
        self.assertEqual(len(self.scheds), 2)
        for _ in iterate_timeout(10, "until priming is complete"):
            state_one = first.sched.local_layout_state.get("tenant-one")
            if state_one:
                break

        for _ in iterate_timeout(
                10, "all schedulers to have the same layout state"):
            if (second.sched.local_layout_state.get(
                    "tenant-one") == state_one):
                break

        self.model_test_component_info.model_api = 32
        with first.sched.layout_update_lock, first.sched.run_handler_lock:
            self.fake_gerrit.addEvent(A.addApproval("Approved", 1))
            self.waitUntilSettled(matcher=[second])

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertEqual(A.data["status"], "MERGED")
        self.assertEqual(B.data["status"], "MERGED")


class TestSemaphoreReleaseUpgrade(ZuulTestCase):
    tenant_config_file = 'config/global-semaphores/main.yaml'

    @model_version(32)
    def test_model_32(self):
        # This tests that a job finishing in one tenant will correctly
        # start a job in another tenant waiting on the semaphore.
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([])
        self.assertBuilds([
            dict(name='test-global-semaphore', changes='1,1'),
        ])

        # Block tenant management event queues so we know that the
        # semaphore release events are dispatched via the pipeline
        # trigger event queue.
        with (management_queue_lock(self.zk_client, "tenant-one"),
              management_queue_lock(self.zk_client, "tenant-two")):

            self.executor_server.hold_jobs_in_build = False
            self.executor_server.release()
            self.waitUntilSettled()

            self.assertHistory([
                dict(name='test-global-semaphore',
                     result='SUCCESS', changes='1,1'),
                dict(name='test-global-semaphore',
                     result='SUCCESS', changes='2,1'),
            ], ordered=False)


class TestOidcSecretSupport(ZuulTestCase):
    tenant_config_file = 'config/secrets/main.yaml'

    @model_version(34)
    def test_model_34(self):
        self._run_test()

    @model_version(35)
    def test_model_35(self):
        self._run_test()

    def _run_test(self):
        with open(os.path.join(FIXTURE_DIR,
                               'config/secrets/git/',
                               'org_project2/zuul-secret.yaml')) as f:
            config = f.read()
        file_dict = {'zuul.yaml': config}

        A = self.fake_gerrit.addFakeChange('org/project2', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(A.reported, 1, "A should report success")
        self.assertHistory([
            dict(name='project2-secret', result='SUCCESS', changes='1,1'),
        ])



# Copyright 2018, 2021 Red Hat
# Copyright 2023, 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import logging

import kubernetes

from zuul.driver.kubernetes.kubernetesmodel import (
    KubernetesInstance,
    KubernetesResource,
)
from zuul.model import QuotaInformation
from zuul.provider import (
    BaseProviderEndpoint,
    statemachine
)


class KubernetesDeleteStateMachine(statemachine.StateMachine):
    NAMESPACE_DELETING = 'deleting namespace'

    def __init__(self, endpoint, node, log):
        self.log = log
        self.node = node
        self.endpoint = endpoint
        self.use_openshift_projects =\
            endpoint.connection.use_openshift_projects
        super().__init__(node.delete_state)

    def advance(self):
        if self.state == self.START:
            if self.node.kubernetes_namespace_id:
                if self.use_openshift_projects:
                    self.endpoint._deleteProject(
                        self.node.kubernetes_namespace_id)
                else:
                    self.endpoint._deleteNamespace(
                        self.node.kubernetes_namespace_id)
                self.state = self.NAMESPACE_DELETING
            else:
                self.state = self.COMPLETE

        # We don't currently wait
        if self.state == self.NAMESPACE_DELETING:
            self.state = self.COMPLETE

        if self.state == self.COMPLETE:
            self.complete = True


class KubernetesCreateStateMachine(statemachine.StateMachine):
    NAMESPACE_CREATING = 'creating namespace'
    SA_CREATING = 'creating service account'
    RBAC_CREATING = 'creating rbac'
    POD_CREATING = 'creating pod'

    def __init__(self, endpoint, node, hostname, label, flavor, image,
                 log):
        self.log = log
        self.endpoint = endpoint
        self.node = node
        self.hostname = hostname
        super().__init__(node.create_state)
        self.label = label
        self.use_openshift_projects =\
            endpoint.connection.use_openshift_projects
        if label.kind == 'pod':
            self.create_pod = True
            self.restricted_access = True
            self.containers = [c['name'] for c in label.spec['containers']]
        elif label.kind == 'namespace':
            self.create_pod = False
            self.restricted_access = False
            self.containers = None
        self.quota = self.endpoint.getQuotaForLabel(label)
        self.user = 'zuul-worker'
        self.namespace = self.hostname
        self.node.kubernetes_connection['user'] = self.user
        self.node.kubernetes_connection['host'] =\
            self.endpoint.core_client.api_client.configuration.host
        self.node.kubernetes_connection['skiptls'] =\
            not self.endpoint.core_client.api_client.configuration.verify_ssl
        self.node.kubernetes_connection['namespace'] = self.namespace
        # Pod and namespace names are the same
        self.node.kubernetes_connection['pod'] = self.namespace
        self.node.kubernetes_connection['containers'] = self.containers

    def advance(self):
        if self.state == self.START:
            if self.use_openshift_projects:
                self.endpoint._createProject(self.hostname, self.node.tags)
            else:
                self.endpoint._createNamespace(self.hostname, self.node.tags)
            self.endpoint._createImagePullSecrets(self.namespace, self.label)
            self.endpoint._createServiceAccount(self.namespace,
                                                self.user)
            self.node.kubernetes_namespace_id = self.namespace
            self.state = self.SA_CREATING

        if self.state == self.SA_CREATING:
            token = self.endpoint._getToken(self.namespace, self.user)
            if token is None:
                return
            (token, ca_crt) = token
            self.node.kubernetes_connection['token'] = token
            if not self.node.kubernetes_connection['skiptls']:
                self.node.kubernetes_connection['ca_crt'] = ca_crt
            self.endpoint._createRole(self.namespace, self.user,
                                      self.restricted_access)
            self.state = self.RBAC_CREATING

        if self.state == self.RBAC_CREATING:
            self.endpoint._createPod(self.namespace, self.hostname,
                                     self.label, self.node.tags)
            self.state = self.POD_CREATING

        if self.state == self.POD_CREATING:
            if not self.endpoint._getPod(self.namespace, self.hostname):
                return
            self.state = self.COMPLETE

        if self.state == self.COMPLETE:
            self.complete = True
            return KubernetesInstance(self.label.kind, self.hostname,
                                      self.node.tags, self.quota)


class KubernetesProviderEndpoint(BaseProviderEndpoint):
    def __init__(self, zk_client, driver, connection, system_id):
        name = connection.connection_name
        super().__init__(zk_client, driver, connection, name, system_id)
        self.log = logging.getLogger(f"zuul.kubernetes.{self.name}")

    def startEndpoint(self):
        self.log.debug("Starting kubernetes endpoint")
        self.core_client, self.rbac_client, self.dynamic_client =\
            self._getClient()
        self._running = True

    def stopEndpoint(self):
        self.log.debug("Stopping kubernetes endpoint")
        self._running = False

    def postConfig(self, provider):
        pass

    def refreshQuotaLimits(self, update):
        if self.quota_cache.hasLimits() and not update:
            return False
        limits = QuotaInformation()
        self.quota_cache.setLimits(limits)
        return True

    def listResources(self, providers):
        if self.connection.use_openshift_projects:
            projects = self.dynamic_client.resources.get(
                api_version='v1', kind='Project')
            namespaces = projects.get().items
            namespace_type = KubernetesResource.TYPE_PROJECT
        else:
            namespaces = self.core_client.list_namespace().items
            namespace_type = KubernetesResource.TYPE_NAMESPACE

        for namespace in namespaces:
            yield KubernetesResource(namespace.metadata.labels or {},
                                     namespace_type,
                                     namespace.metadata.name)

    def deleteResource(self, resource):
        self.log.info(f"Deleting leaked {resource.type}: {resource.id}")
        if resource.type == KubernetesResource.TYPE_NAMESPACE:
            self._deleteNamespace(resource.id)
        elif resource.type == KubernetesResource.TYPE_PROJECT:
            self._deleteProject(resource.id)

    def listInstances(self):
        if self.connection.use_openshift_projects:
            projects = self.dynamic_client.resources.get(
                api_version='v1', kind='Project')
            namespaces = projects.get().items
        else:
            namespaces = self.core_client.list_namespace().items

        for namespace in namespaces:
            quota = QuotaInformation(namespaces=1)
            yield KubernetesInstance('namespace',
                                     namespace.metadata.name,
                                     namespace.metadata.labels or {},
                                     quota)
            for pod in self.core_client.list_namespaced_pod(
                    namespace.metadata.name).items:
                quota = QuotaInformation(pods=1)
                yield KubernetesInstance('pod',
                                         pod.metadata.name,
                                         pod.metadata.labels or {},
                                         quota)

    # Local implementation
    def getQuotaForLabel(self, label):
        resources = {'namespaces': 1}
        if label.kind == 'pod':
            resources['pods'] = 1
        return QuotaInformation(**resources)

    def _getConfig(self, config_file, context):
        try:
            return kubernetes.config.new_client_from_config(
                config_file=config_file, context=context)
        except FileNotFoundError:
            self.log.debug("Kubernetes config file not found, attempting "
                           "to load in-cluster configs")
            return kubernetes.config.load_incluster_config()
        except kubernetes.config.config_exception.ConfigException as e:
            if 'Invalid kube-config file. No configuration found.' in str(e):
                self.log.debug("Kubernetes config file not found, attempting "
                               "to load in-cluster configs")
                return kubernetes.config.load_incluster_config()
            else:
                raise

    def _getClient(self):
        config_file = self.connection.config_file
        context = self.connection.context
        conf = self._getConfig(config_file, context)
        core_client = kubernetes.client.CoreV1Api(conf)
        rbac_client = kubernetes.client.RbacAuthorizationV1Api(conf)
        api_client = kubernetes.client.api_client.ApiClient(conf)
        dynamic_client = kubernetes.dynamic.DynamicClient(api_client)
        return (core_client, rbac_client, dynamic_client)

    def _createNamespace(self, namespace, labels):
        self.log.debug("Creating namespace %s", namespace)

        # Create the namespace
        ns_body = {
            'apiVersion': 'v1',
            'kind': 'Namespace',
            'metadata': {
                'name': namespace,
                'labels': labels,
            }
        }
        self.core_client.create_namespace(ns_body)

    def _createProject(self, project, labels):
        self.log.debug("Creating project %s", project)

        # Create the project
        proj_body = {
            'apiVersion': 'project.openshift.io/v1',
            'kind': 'ProjectRequest',
            'metadata': {
                'name': project,
                'labels': labels,
            }
        }
        projects = self.dynamic_client.resources.get(
            api_version='project.openshift.io/v1', kind='ProjectRequest')
        projects.create(body=proj_body)

    def _createImagePullSecrets(self, namespace, label):
        # Copy any image pull secrets required
        for conf_secret in label.image_pull_secrets:
            old_secret = self.core_client.read_namespaced_secret(
                conf_secret['name'], conf_secret['namespace'])
            new_secret = {
                'apiVersion': 'v1',
                'kind': 'Secret',
                'type': 'kubernetes.io/dockerconfigjson',
                'metadata': {
                    'name': conf_secret['name'],
                },
                'data': old_secret.data,
            }
            self.core_client.create_namespaced_secret(namespace, new_secret)

    def _createServiceAccount(self, namespace, user):
        sa_body = {
            'apiVersion': 'v1',
            'kind': 'ServiceAccount',
            'metadata': {'name': user}
        }
        self.core_client.create_namespaced_service_account(namespace, sa_body)

        secret_body = {
            'apiVersion': 'v1',
            'kind': 'Secret',
            'type': 'kubernetes.io/service-account-token',
            'metadata': {
                'name': user,
                'annotations': {
                    'kubernetes.io/service-account.name': user
                }
            }
        }
        self.core_client.create_namespaced_secret(namespace, secret_body)

    def _getToken(self, namespace, user):
        secret = self.core_client.read_namespaced_secret(user, namespace)
        ca_crt = None
        token = None
        if secret.data:
            token = secret.data.get('token')
            ca_crt = secret.data.get('ca.crt')
            if token and ca_crt:
                token = base64.b64decode(
                    token.encode('utf-8')).decode('utf-8')
                return (token, ca_crt)

    def _createRole(self, namespace, user, restricted_access):
        # Create service account role
        all_verbs = ["create", "delete", "get", "list", "patch",
                     "update", "watch"]
        if restricted_access:
            role_name = "zuul-restricted"
            role_body = {
                'kind': 'Role',
                'apiVersion': 'rbac.authorization.k8s.io/v1',
                'metadata': {
                    'name': role_name,
                },
                'rules': [{
                    'apiGroups': [""],
                    'resources': ["pods"],
                    'verbs': ["get", "list"],
                }, {
                    'apiGroups': [""],
                    'resources': ["pods/exec"],
                    'verbs': all_verbs
                }, {
                    'apiGroups': [""],
                    'resources': ["pods/logs"],
                    'verbs': all_verbs
                }, {
                    'apiGroups': [""],
                    'resources': ["pods/portforward"],
                    'verbs': all_verbs
                }]
            }
        else:
            role_name = "zuul"
            role_body = {
                'kind': 'Role',
                'apiVersion': 'rbac.authorization.k8s.io/v1',
                'metadata': {
                    'name': role_name,
                },
                'rules': [{
                    'apiGroups': ["apps"],
                    'resources': ["deployments", "replicasets"],
                    'verbs': all_verbs,
                }, {
                    'apiGroups': ["batch"],
                    'resources': ["cronjobs", "jobs"],
                    'verbs': all_verbs,
                }, {
                    'apiGroups': [""],
                    'resources': ["pods", "pods/exec", "pods/log",
                                  "pods/portforward", "services",
                                  "endpoints", "configmaps", "secrets"],
                    'verbs': all_verbs,
                }]
            }
        self.rbac_client.create_namespaced_role(namespace, role_body)

        # Give service account admin access
        role_binding_body = {
            'apiVersion': 'rbac.authorization.k8s.io/v1',
            'kind': 'RoleBinding',
            'metadata': {'name': 'zuul-role'},
            'roleRef': {
                'apiGroup': 'rbac.authorization.k8s.io',
                'kind': 'Role',
                'name': role_name,
            },
            'subjects': [{
                'kind': 'ServiceAccount',
                'name': user,
                'namespace': namespace,
            }],
            'userNames': ['system:serviceaccount:%s:zuul-worker' % namespace]
        }
        self.rbac_client.create_namespaced_role_binding(
            namespace, role_binding_body)

    def _createPod(self, namespace, name, label, tags):
        kubernetes_labels = tags
        kubernetes_annotations = {}
        if label.annotations:
            kubernetes_annotations.update(label.annotations)

        pod_body = {
            'apiVersion': 'v1',
            'kind': 'Pod',
            'metadata': {
                'name': name,
                'labels': kubernetes_labels,
                'annotations': kubernetes_annotations,
            },
            'spec': label.spec,
            'restartPolicy': 'Never',
        }

        self.core_client.create_namespaced_pod(namespace, pod_body)

    def _getPod(self, namespace, name):
        pod = self.core_client.read_namespaced_pod(name, namespace)
        if pod.status.phase == "Running":
            return True
        return False

    def _deleteNamespace(self, namespace):
        self.log.debug("Deleting namespace %s", namespace)

        delete_body = {
            "apiVersion": "v1",
            "kind": "DeleteOptions",
            "propagationPolicy": "Background"
        }
        self.core_client.delete_namespace(namespace, body=delete_body)

    def _deleteProject(self, project):
        self.log.debug("Deleting project %s", project)
        projects = self.dynamic_client.resources.get(
            api_version='v1', kind='Project')
        projects.delete(name=project)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from kazoo.exceptions import KazooException


class ZuulZooKeeperException(KazooException):
    """Base exception class for all custom ZK exceptions"""
    pass


class LockException(ZuulZooKeeperException):
    pass


class NoClientException(ZuulZooKeeperException):

    def __init__(self):
        super().__init__("No zookeeper client!")


class JobRequestNotFound(ZuulZooKeeperException):
    pass
